/* This is the declaration of a class to manage plugins 

   There are three main groups of functions:
   * those to load plugins;
   * those to query the manager for the list of devices it supports;
   * those to get a pointer to an already initialized object;

   The interface shows the dichotomy between two main usages of plugins:
   * those that come with pwlib, upon which the client program should know
   little, and fully rely on the manager to get them;
   * those that come with the client program, that the client program may want
   to have a more direct access to;

   The exact details of how to code a plugin depend on the implementation of
   that class; see src/ptlib/unix/plugins_dlopen.cxx for example.

   The plugins fall in several categories, depending on whether they manage
   devices able to do video none/input/output/both and/or sound with the same
   different capabilities.

   There are two types of names to get to a real device:
   * full name "DRIVER SHORTNAME"; the manager strips the driver name and gives
   the short name to the corresponding driver; that allows a single SHORTNAME
   to be handled by several drivers;
   * short name, for example "/dev/video3.14" or "PiCam"; this is the name that
   only plugins are supposed to understand; as such, the manager should try
   to scan the plugins for one that accepts it, and use the first that does.

*/

#include <ptlib.h>
#include <ptlib/videoio.h>

class PDeviceManager
{
public:
  /* those functions should never get called: the typical use of the manager
     is through a direct call to the static member functions */
  PDeviceManager();
  ~PDeviceManager();

  /* This is the plugin type, that allows them to declare what they are
     able to do
  */
  enum plugin_type {
    VIDIN = 1,
    VIDOUT = 2,
    SNDIN = 4,
    SNDOUT = 8
  };

  enum Direction { INPUT, OUTPUT };

  /* These functions are there to load plugins; either by pointing the manager
     to a directory where they are to be found, or directly to the filename.
     
     LoadPlugin returns TRUE if the plugin could be loaded and FALSE otherwise
     LoadPluginDirectory returns FALSE only if no plugin could be loaded
  */
  static BOOL LoadPluginDirectory(PDirectory directory_name);
  static BOOL LoadPlugin(PString plugin_filename);

  /* These functions will return the list of all supported devices,
     in a correctly mangled way: for example, for video input,
     "V4L /dev/video3.14"; the first part being the plugin name "V4L"
     that handles the device, and "/dev/video3.14" is either directly
     a filename, or a name that the plugin will know how to use anyway,
     like a more userfriendly "PiCam", that the plugin will then relate to
     "/dev/video3.14".
  */
  static PStringList GetVideoDeviceNames(Direction);
  static PStringArray GetSoundDeviceNames(Direction);

  /* Those two functions are the ones used to get a pointer to an usual,
     already created PVideoInputDevice/PSoundChannel (ie: equivalent to a
     'new PVideoInputDevice/PSoundChannel').

     Get*Device should be used in case the program already asked for a plugin
     through LoadPlugin, and wants direct access to it; it returns a pointer to
     a non-opened PVideoInputDevice/PSoundChannel, since it is assumed the
     program knows the plugin, and how to directly talk with it;

     GetOpened*Device should be used in case the program has no particular
     knowledge about the plugin; as such it takes the full device name (ie:
     "V4L PiCam"), and the same other arguments than the Open call in the
     corresponding class, then return a pointer to an ALREADY OPENED 
     PVideoInputDevice/PSoundChannel.
   */  
  static PVideoInputDevice *GetVideoInputDevice(PString plugin_name);
  static PVideoInputDevice *GetOpenedVideoInputDevice(PString full_name,
						      BOOL startImmediate);
  static PVideoOutputDevice *GetVideoOutputDevice(PString plugin_name);
  static PVideoOutputDevice *GetOpenedVideoOutputDevice(PString full_name,
						      BOOL startImmediate);
  static PSoundChannel *GetSoundDevice(PString plugin_name, Direction);
  static PSoundChannel *GetOpenedSoundDevice(PString full_name,
					     Direction,
					     unsigned numChannels,
					     unsigned sampleRate,
					     unsigned bitsPerSample);

};

#define DECLARE_PLUGIN(NAME, TYPE) \
extern "C" char *plugin_getname(void) { return NAME; }; \
extern "C" int plugin_gettype(void) { return TYPE; };

#define DECLARE_PLUGIN_VIDEOINPUT(CLASS) \
extern "C" PVideoInputDevice *plugin_create_vidin() { return new CLASS(); } \
extern "C" PStringList plugin_getdevices_vidin() { return CLASS::GetInputDeviceNames(); }

#define DECLARE_PLUGIN_VIDEOOUTPUT(CLASS) \
extern "C" PVideoOutputDevice *plugin_create_vidout() { return new CLASS(); } \
extern "C" PStringList plugin_getdevices_vidout() { return CLASS::GetDeviceNames(); }

#define DECLARE_PLUGIN_SOUNDINPUT(CLASS) \
extern "C" PSoundChannel *plugin_create_sndin() { return new CLASS(); } \
extern "C" PStringArray plugin_getdevices_sndin() { return CLASS::GetDeviceNames(PSoundChannel::Recorder); }

#define DECLARE_PLUGIN_SOUNDOUTPUT(CLASS) \
extern "C" PSoundChannel *plugin_create_sndout() { return new CLASS(); } \
extern "C" PStringArray plugin_getdevices_sndout() { return CLASS::GetDeviceNames(PSoundChannel::Player); }
